﻿//License: GPLv3 (see LICENSE.txt)

using System.Collections.Generic;
using System.Collections;
using System.Linq;
using System;
using UnityEngine;

namespace Rotate
{
    [KSPAddon(KSPAddon.Startup.Flight, false)]
    public class VehicleRotator : MonoBehaviour
    {
        //GENERAL VARIABLES
        Dictionary<string, Vector3> lastPosition = new Dictionary<string, Vector3>();
        Dictionary<string, Transform> lastTransform = new Dictionary<string, Transform>();
        Dictionary<string, bool> lastActive = new Dictionary<string, bool>();
        Dictionary<string, ITargetable> lastReference = new Dictionary<string, ITargetable>();
        Vessel lastVessel;

        // +++ CFG & DICTIONARYS +++
        Dictionary<string, Vector3> vessel_momentum = new Dictionary<string, Vector3>();
        Dictionary<string, Quaternion> vessel_rotation = new Dictionary<string, Quaternion>();
        Dictionary<string, Vector3> vessel_direction = new Dictionary<string, Vector3>();
        Dictionary<string, ITargetable> vessel_reference = new Dictionary<string, ITargetable>();
        String rot_cfg
        {
            get { return KSPUtil.ApplicationRootPath + "/GameData/PersistentRotation/PersistentRotation_" + HighLogic.CurrentGame.Title.TrimEnd("_()SANDBOXCAREERSCIENCE".ToCharArray()).TrimEnd(' ') + ".cfg"; }
        }
        String gui_cfg
        {
            get { return KSPUtil.ApplicationRootPath + "/GameData/PersistentRotation/GUIconfig.cfg"; }
        }
        private void CleanCfg()
        {
            //Debug.Log("Clean in");
            //Clean vessel momentum
            Dictionary<string, Vector3> temp_vessel_momentum = new Dictionary<string, Vector3>();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                temp_vessel_momentum[vessel.id.ToString()] = Vector3.zero;
            }
            foreach (KeyValuePair<string, Vector3> entry in vessel_momentum)
            {
                temp_vessel_momentum[entry.Key] = entry.Value;
            }
            vessel_momentum.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_momentum[vessel.id.ToString()] = temp_vessel_momentum[vessel.id.ToString()];
            }
            temp_vessel_momentum.Clear();
            //Clean vessel rotation
            Dictionary<string, Quaternion> temp_vessel_rotation = new Dictionary<string, Quaternion>();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                temp_vessel_rotation[vessel.id.ToString()] = Quaternion.identity;
            }
            foreach (KeyValuePair<string, Quaternion> entry in vessel_rotation)
            {
                temp_vessel_rotation[entry.Key] = entry.Value;
            }
            vessel_rotation.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_rotation[vessel.id.ToString()] = temp_vessel_rotation[vessel.id.ToString()];
            }
            temp_vessel_rotation.Clear();
            //Clean vessel direction
            Dictionary<string, Vector3> temp_vessel_direction = new Dictionary<string, Vector3>();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                temp_vessel_direction[vessel.id.ToString()] = Vector3.zero;
            }
            foreach (KeyValuePair<string, Vector3> entry in vessel_direction)
            {
                temp_vessel_direction[entry.Key] = entry.Value;
            }
            vessel_direction.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_direction[vessel.id.ToString()] = temp_vessel_direction[vessel.id.ToString()];
            }
            temp_vessel_direction.Clear();
            //Clean vessel reference
            Dictionary<string, ITargetable> temp_vessel_reference = new Dictionary<string, ITargetable>();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                temp_vessel_reference[vessel.id.ToString()] = null;
            }
            foreach (KeyValuePair<string, ITargetable> entry in vessel_reference)
            {
                temp_vessel_reference[entry.Key] = entry.Value;
            }
            vessel_reference.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_reference[vessel.id.ToString()] = temp_vessel_reference[vessel.id.ToString()];
            }
            temp_vessel_reference.Clear();
            //Debug.Log("Clean out");
        }
        private void Save()
        {
            //Debug.Log("Save in");
            try
            {
                CleanCfg();

                ConfigNode save = new ConfigNode();
                ConfigNode.CreateConfigFromObject(this, 0, save);

                //Save momentum of all vessels
                ConfigNode momentum = save.AddNode("MOMENTUM");
                foreach (KeyValuePair<string, Vector3> entry in vessel_momentum)
                {
                    momentum.AddValue(entry.Key, KSPUtil.WriteVector(entry.Value));
                }
                //Save rotation of all vessels
                ConfigNode rotation = save.AddNode("ROTATION");
                foreach (KeyValuePair<string, Quaternion> entry in vessel_rotation)
                {
                    rotation.AddValue(entry.Key, KSPUtil.WriteQuaternion(entry.Value));
                }
                //Save direction to the mainBody for each vessel
                ConfigNode direction = save.AddNode("DIRECTION");
                foreach (KeyValuePair<string, Vector3> entry in vessel_direction)
                {
                    direction.AddValue(entry.Key, KSPUtil.WriteVector(entry.Value));
                }
                //Save direction reference object for each vessel
                ConfigNode reference = save.AddNode("REFERENCE");
                foreach (KeyValuePair<string, ITargetable> entry in vessel_reference)
                {
                    if (entry.Value != null)
                    {
                        Debug.Log(entry.Value.GetType().ToString());
                        if (entry.Value.GetType() == typeof(CelestialBody))
                        {
                            reference.AddValue(entry.Key, entry.Value.GetName());
                        }
                        else if (entry.Value.GetType() == typeof(Vessel))
                        {
                            reference.AddValue(entry.Key, entry.Value.GetVessel().id.ToString());
                        }
                        else
                        {
                            Debug.LogError("[PR]: Wrong Reference Type!");
                            reference.AddValue(entry.Key, "NONE");
                        }
                    }
                    else
                    {
                        reference.AddValue(entry.Key, "NONE");
                    }
                }
                save.Save(rot_cfg);
            }
            catch (Exception e) { Debug.Log("[PR] Saving not sucessfull: " + e.Message); }

            //Debug.Log("Save out");
        }
        private void Load()
        {
            Debug.Log("Load in");
            //This is called when all persistent rotation data is being loaded from the cfg file.

            float ran_x;
            float ran_y;
            float ran_z;

            //vessel_momentum from scratch
            vessel_momentum.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                if (vessel.vesselType == VesselType.SpaceObject || vessel.vesselType == VesselType.Unknown)
                {
                    //Random momentum for every asteroid if it has no momentum already.
                    ran_x = (float)UnityEngine.Random.Range(0, 10) / (float)100;
                    ran_y = (float)UnityEngine.Random.Range(0, 10) / (float)100;
                    ran_z = (float)UnityEngine.Random.Range(0, 10) / (float)100;

                    vessel_momentum[vessel.id.ToString()] = new Vector3(ran_x, ran_y, ran_z);
                }
                else
                {
                    vessel_momentum[vessel.id.ToString()] = Vector3.zero;
                }
            }

            //vessel_rotation from scratch
            vessel_rotation.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_rotation[vessel.id.ToString()] = vessel.transform.rotation;
            }

            //vessel_direction from scratch
            vessel_direction.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_direction[vessel.id.ToString()] = (vessel.mainBody.position - vessel.transform.position).normalized;
            }

            //vessel_reference from scratch
            vessel_reference.Clear();
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_reference[vessel.id.ToString()] = null;
            }
            //Open cfg file
            ConfigNode load = ConfigNode.Load(rot_cfg);
            if (load == null)
            {
                Debug.Log("[PR] Save file is empty or not existent!");
                return;
            }
            //Load vessel_momentum
            ConfigNode momentum = load.GetNode("MOMENTUM");
            foreach (ConfigNode.Value s in momentum.values)
            {
                vessel_momentum[s.name] = KSPUtil.ParseVector3(s.value);
            }
            //Load vessel_rotation
            ConfigNode rotation = load.GetNode("ROTATION");
            foreach (ConfigNode.Value s in rotation.values)
            {
                vessel_rotation[s.name] = KSPUtil.ParseQuaternion(s.value);
            }
            //Load vessel_direction
            ConfigNode direction = load.GetNode("DIRECTION");
            foreach (ConfigNode.Value s in direction.values)
            {
                vessel_direction[s.name] = KSPUtil.ParseVector3(s.value);
            }
            //Load vessel_reference
            ConfigNode reference = load.GetNode("REFERENCE");
            foreach (ConfigNode.Value s in reference.values)
            {
                if (s.value != "NONE")
                {
                    vessel_reference[s.name] = null;
                    foreach (CelestialBody body in FlightGlobals.Bodies) //Check all bodies
                    {
                            if (body.name == s.value)
                            {
                                vessel_reference[s.name] = body;
                            }
                    }

                    foreach (Vessel vessel in FlightGlobals.Vessels) // Check all vessels
                    {
                        if (vessel.id.ToString() == s.value)
                        {
                            vessel_reference[s.name] = vessel;
                        }
                    }
                }
                else
                {
                    vessel_reference[s.name] = null;
                }
            }
            Debug.Log("Load out");
        }

        // +++ GAME EVENTS +++
        void OnVesselGoOnRails(Vessel vessel)
        {
            //Debug.Log("VesselOnRails in");
            if (vessel.ActionGroups[KSPActionGroup.SAS])
            {
                //If Autopilot is enabled when vessel goes on rails
                vessel_momentum[vessel.id.ToString()] = Vector3.zero;
            }
            else
            {
                //NOTE: Issue with non-existent ReferenceTransform seems fixed

                if (lastVessel.ReferenceTransform) //If the vessel has a ReferenceTransform part
                {
                    vessel_momentum[vessel.id.ToString()] = Quaternion.Inverse(Quaternion.Inverse(lastVessel.ReferenceTransform.rotation) * lastVessel.transform.rotation) * vessel.angularVelocity;
                }
                else //Kerbals, Flags etc.
                {
                    Debug.LogWarning("No ReferenceTransform found in vessel " + lastVessel.GetName() + ", using fallback method.");
                    vessel_momentum[vessel.id.ToString()] = vessel.angularVelocity;
                }
            }

            vessel_rotation[vessel.id.ToString()] =  vessel.transform.rotation;

            if (vessel_reference.ContainsKey(vessel.id.ToString()))
            {
                if (vessel_reference[vessel.id.ToString()] != null)
                {
                    vessel_direction[vessel.id.ToString()] = vessel_reference[vessel.id.ToString()].GetTransform().position - vessel.transform.position;
                }
            }
            else
            {
                Debug.Log("[PR] No key! 1");
            }
            Save();
            //Debug.Log("VesselOnRails out");
        }
        void OnVesselGoOffRails(Vessel vessel)
        {
            //Debug.Log("VesselOffRails in");
            Load();
            CleanCfg();
            vessel.Autopilot.SAS.lockedHeading = vessel.ReferenceTransform.rotation;

            if (vessel.ActionGroups[KSPActionGroup.SAS])
            {
                if (vessel_reference.ContainsKey(vessel.id.ToString()) && vessel_direction.ContainsKey(vessel.id.ToString()) && vessel_rotation.ContainsKey(vessel.id.ToString()))
                {
                    if (vessel_reference[vessel.id.ToString()] != null)
                    {

                        vessel.SetRotation(Quaternion.FromToRotation(vessel_direction[vessel.id.ToString()], vessel_reference[vessel.id.ToString()].GetTransform().position - vessel.transform.position) * vessel_rotation[vessel.id.ToString()]);
                    }
                }
                else
                {
                    Debug.Log("[PR] No key! 2");
                }
            }
            else
            {
                if(vessel_momentum.ContainsKey(vessel.id.ToString()))
                {
                    Vector3 av = vessel_momentum[vessel.id.ToString()];
                    Vector3 COM = vessel.findWorldCenterOfMass();
                    foreach (Part p in vessel.parts) //Applying force on every part
                    {
                        try
                        {
                            if (p.rigidbody == null) continue;
                            p.rigidbody.AddTorque(vessel.transform.rotation * av, ForceMode.VelocityChange);
                            p.rigidbody.AddForce(Vector3.Cross(vessel.transform.rotation * av, (p.rigidbody.position - COM)), ForceMode.VelocityChange);
                            p.rigidbody.angularDrag = 0.0f; //Testing
                        }
                        catch (NullReferenceException nre)
                        {
                            Debug.Log("[PR] NullReferenceException in OnVesselGoOffRails: " + nre.Message);
                        }
                    }
                }
                else
                {
                    Debug.Log("[PR] No key! 3");
                }
            }
            //Debug.Log("VesselOffRails out");
        }
        void Awake()
        {
            //Debug.Log("Awake in");
            //Add GameEvents
            GameEvents.onVesselGoOnRails.Add(OnVesselGoOnRails);
            GameEvents.onVesselGoOffRails.Add(OnVesselGoOffRails);
            GameEvents.onGamePause.Add(OnPause);
            GameEvents.onShowUI.Add(OnShowUI);
            GameEvents.onHideUI.Add(OnHideUI);

            //Initialize GUI Stuff
            mainGuid = Guid.NewGuid().GetHashCode();
            bodyGuid = Guid.NewGuid().GetHashCode();
            configGuid = Guid.NewGuid().GetHashCode();
            gravityGuid = Guid.NewGuid().GetHashCode();

            MainWindowRect = new Rect(Screen.width / 4, 0, 200, 10); //Overwritten by LoadGUI
            BodyWindowRect = new Rect((Screen.width / 2) - 75, Screen.height / 4, 150, 10);
            ConfigWindowRect = new Rect((Screen.width / 2) - 100, Screen.height / 4, 200, 10);
            GravityWindowRect = new Rect((Screen.width / 2) - 100, Screen.height / 4, 200, 10);

            LoadGUI();

            if (visibility_mode == 2)
            {
                CreateStockToolbar();
            }
            else if (visibility_mode == 3)
            {
                CreateBlizzyToolbar();
            }

            //Debug.Log("Awake out");
        }
        void Start()
        {
            // Add local functions to Game Events
            //Debug.Log("Start in");
            Load();
            CleanCfg();

            foreach(Vessel vessel in FlightGlobals.Vessels)
            {
                lastTransform[vessel.id.ToString()] = vessel.ReferenceTransform;
                if (lastTransform.ContainsKey(vessel.id.ToString()) && vessel_reference.ContainsKey(vessel.id.ToString()))
                {
                    if (vessel_reference[vessel.id.ToString()] != null)
                    {
                        lastPosition[vessel.id.ToString()] = (Vector3d)lastTransform[vessel.id.ToString()].position - vessel_reference[vessel.id.ToString()].GetTransform().position;
                    }
                }
                else
                {
                    Debug.Log("[PR] No key! 4");
                }
                
                lastActive[vessel.id.ToString()] = false;
            }
            //Debug.Log("Start out");
        }
        void OnPause()
        {
            //Debug.Log("OnPause in");
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                if (!vessel.packed)
                {
                    vessel_momentum[vessel.id.ToString()] = vessel.angularVelocity;
                    vessel_rotation[vessel.id.ToString()] = vessel.transform.rotation;
                    if (vessel_reference.ContainsKey(vessel.id.ToString()))
                    {
                        if (vessel_reference[vessel.id.ToString()] != null)
                        {
                            vessel_direction[vessel.id.ToString()] = (vessel_reference[vessel.id.ToString()].GetTransform().position - vessel.transform.position).normalized;
                        }
                    }
                    else
                    {
                        Debug.Log("[PR] No key! 5");
                    }
                }
            }
            Save();
            //Debug.Log("OnPause out");
        }
        void OnDestroy()
        {
            SaveGUI();

            DeleteBlizzyToolbar();
            DeleteStockToolbar();

            //Unbind functions from GameEvents
            GameEvents.onVesselGoOnRails.Remove(OnVesselGoOnRails);
            GameEvents.onVesselGoOffRails.Remove(OnVesselGoOffRails);
            GameEvents.onGamePause.Remove(OnPause);
            GameEvents.onShowUI.Remove(OnShowUI);
            GameEvents.onHideUI.Remove(OnHideUI);
        }

        // +++ PHYSICS +++
        private void SurfaceSAS(Vessel vessel)
        {
            if (vessel.Autopilot.Mode == VesselAutopilot.AutopilotMode.StabilityAssist)
            {
                if (lastReference.ContainsKey(vessel.id.ToString()) && lastActive.ContainsKey(vessel.id.ToString()) && lastTransform.ContainsKey(vessel.id.ToString()) && vessel_reference.ContainsKey(vessel.id.ToString()) && lastPosition.ContainsKey(vessel.id.ToString()))
                {
                    if (vessel_reference[vessel.id.ToString()] != null) //Check vessel_reference in current frame
                    {
                        if (lastActive[vessel.id.ToString()])
                        {
                            if (lastReference[vessel.id.ToString()] == vessel_reference[vessel.id.ToString()])
                            {
                                Vector3d newPosition = (Vector3d)lastTransform[vessel.id.ToString()].position - vessel_reference[vessel.id.ToString()].GetTransform().position;
                                QuaternionD delta = FromToRotation(lastPosition[vessel.id.ToString()], newPosition);
                                QuaternionD adjusted = delta * (QuaternionD)vessel.Autopilot.SAS.lockedHeading;
                                vessel.Autopilot.SAS.lockedHeading = adjusted;
                            }
                        }
                    }
                }
                else
                {
                    Debug.Log("[PR] No key! 6");
                }
            }
        }
        private void RotateWhilePacked(Vessel vessel)
        {
            float ran_x, ran_y, ran_z;
            //Debug.Log("Rotate in");
            Vector3 angvel;
            if (!vessel_momentum.ContainsKey(vessel.id.ToString()))
            {
                 //Random momentum for every asteroid if it has no momentum already.
                 if (vessel.vesselType == VesselType.SpaceObject || vessel.vesselType == VesselType.Unknown)
                 {
                        ran_x = (float)UnityEngine.Random.Range(0, 10) / (float)100;
                        ran_y = (float)UnityEngine.Random.Range(0, 10) / (float)100;
                        ran_z = (float)UnityEngine.Random.Range(0, 10) / (float)100;

                        vessel_momentum[vessel.id.ToString()] = new Vector3(ran_x, ran_y, ran_z);

                        Debug.Log("Random Rotation for Asteroid " + vessel.vesselName + " initiated with magnitude of " + vessel_momentum[vessel.id.ToString()].magnitude);
                }
                else
                {
                    vessel_momentum[vessel.id.ToString()] = Vector3.zero;
                }

                Debug.Log("[PR] No key! 7");
            }

            angvel = vessel_momentum[vessel.id.ToString()];

            if (vessel.situation != Vessel.Situations.LANDED
                && vessel.situation != Vessel.Situations.PRELAUNCH
                && vessel.situation != Vessel.Situations.DOCKED
                && vessel.situation != Vessel.Situations.SPLASHED)
            {
                //Either the vessel continues to rotate, or it follows the planets curvature. This might change.
                if (vessel.ActionGroups[KSPActionGroup.SAS])
                {
                    if (vessel.Autopilot.Mode == VesselAutopilot.AutopilotMode.StabilityAssist)
                    {
                        if (lastReference.ContainsKey(vessel.id.ToString()) && vessel_direction.ContainsKey(vessel.id.ToString()) && vessel_reference.ContainsKey(vessel.id.ToString()) && vessel_rotation.ContainsKey(vessel.id.ToString()))
                        {
                            if (vessel_reference[vessel.id.ToString()] != null)
                            {
                                if (lastReference[vessel.id.ToString()] == vessel_reference[vessel.id.ToString()])
                                {
                                    vessel.SetRotation(Quaternion.FromToRotation(vessel_direction[vessel.id.ToString()], vessel_reference[vessel.id.ToString()].GetTransform().position - vessel.transform.position) * vessel_rotation[vessel.id.ToString()]);
                                }
                            }
                        }
                        else
                        {
                            Debug.Log("[PR] No key! 8");
                        }
                    }
                }
                else
                {
                    float angle = angvel.magnitude * TimeWarp.CurrentRate;
                    Quaternion rotation = Quaternion.AngleAxis(angle, angvel);
                    vessel.SetRotation(vessel.transform.rotation * rotation);
                }
            }
            //Debug.Log("Rotate out");
        }
        Dictionary<ProtoCrewMember, float> ArtificialGravity = new Dictionary<ProtoCrewMember, float>();
        private void CalculateArtificialGeForce()
        {
            ArtificialGravity.Clear();
            try
            {
                if (FlightGlobals.ActiveVessel.GetCrewCount() > 0)
                {
                    foreach (ProtoCrewMember kerb in FlightGlobals.ActiveVessel.GetVesselCrew())
                    {
                        if (kerb.seat)
                        {
                            float speed = FlightGlobals.ActiveVessel.angularVelocity.magnitude * (kerb.seat.part.rigidbody.position - FlightGlobals.ActiveVessel.findWorldCenterOfMass()).magnitude;
                            float force = (speed * speed) / (kerb.seat.part.rigidbody.position - FlightGlobals.ActiveVessel.findWorldCenterOfMass()).magnitude;
                            ArtificialGravity[kerb] = force;
                        }
                    }
                }
            }
            catch (NullReferenceException e) { Debug.Log("[PR] NullReferenceException in CalculateArtificialGeForce: " + e.Message); }
        }

        // +++ TIMER +++
        int ticker = 0;
        int old_crew_count = 0;
        private void FixedUpdate()
        {
            if (FlightGlobals.ActiveVessel != lastVessel)
                CleanCfg();

            if(old_crew_count != FlightGlobals.ActiveVessel.GetCrewCount())
            {
                GravityWindowRect.height = 10; //Reset height
            }
            old_crew_count = FlightGlobals.ActiveVessel.GetCrewCount();

            ticker++;
            if (ticker > 100)
            {
                ticker = 0;
                CleanCfg();
                SaveGUI();
            }

            CalculateArtificialGeForce();

            try
            {
                foreach (Vessel vessel in FlightGlobals.Vessels)
                {
                    if (vessel.packed)
                    {
                        RotateWhilePacked(vessel);
                        lastActive[vessel.id.ToString()] = false;
                    }
                    else
                    {
                        SurfaceSAS(vessel);

                        if(vessel.ActionGroups[KSPActionGroup.SAS])
                        {
                            lastActive[vessel.id.ToString()] = true;
                        }
                        else
                        {
                            lastActive[vessel.id.ToString()] = false;
                        }

                        lastTransform[vessel.id.ToString()] = vessel.ReferenceTransform;

                        if (vessel_reference.ContainsKey(vessel.id.ToString()))
                        {
                            if (vessel_reference[vessel.id.ToString()] != null)
                            {
                                lastPosition[vessel.id.ToString()] = (Vector3d)lastTransform[vessel.id.ToString()].position - vessel_reference[vessel.id.ToString()].GetTransform().position;
                            }
                            else
                            {
                                lastPosition[vessel.id.ToString()] = Vector3.zero;
                            }
                        }
                        else
                        {
                            vessel_reference[vessel.id.ToString()] = null;
                        }
                    }

                    lastReference[vessel.id.ToString()] = vessel_reference[vessel.id.ToString()];
                }
            }
            catch (NullReferenceException nre)
            {
                Debug.Log("[PR] NullReferenceException in FixedUpdate: " + nre.ToString());
            }

            lastVessel = FlightGlobals.ActiveVessel;
        }

        // +++ QUATERNIONS +++
        public QuaternionD FromToRotation(Vector3d fromv, Vector3d tov) //Stock FromToRotation() doesn't work correctly
        {
            Vector3d cross = Vector3d.Cross(fromv, tov);
            double dot = Vector3d.Dot(fromv, tov);
            double wval = dot + Math.Sqrt(fromv.sqrMagnitude * tov.sqrMagnitude);
            double norm = 1.0 / Math.Sqrt(cross.sqrMagnitude + wval * wval);
            return new QuaternionD(cross.x * norm, cross.y * norm, cross.z * norm, wval * norm);
        }

        // +++ GRAPHICAL USER INTERFACE +++
        private IButton button;
        private ApplicationLauncherButton stockButton;
        int mainGuid;
        int bodyGuid;
        int configGuid;
        int gravityGuid;
        Rect BodyWindowRect;
        Rect MainWindowRect;
        Rect ConfigWindowRect;
        Rect GravityWindowRect;
        bool hidden = false; //If F2 is pressed
        bool disabled = false; //If Toolbar is pressed, overwritten by LoadGUI
        bool showMainWindow = true; //Window visible, overwritten by LoadGUI
        bool showBodyWindow = false;
        bool showConfigWindow = false;
        bool showGravityWindow = false;
        bool MainWindowActive = false; // minimize-maximize, overwritten by LoadGUI
        int visibility_mode = 1; // 1: Always Visible, 2: Stock Toolbar, 3: Blizzys Toolbar, overwritten by LoadGUI
        Texture close = GameDatabase.Instance.GetTexture("PersistentRotation/Textures/close_w", false);
        Texture options = GameDatabase.Instance.GetTexture("PersistentRotation/Textures/options_w", false);

        void OnGUI()
        {
            if (!hidden)
            {
                if (!disabled)
                {
                    if (showMainWindow)
                    {
                        MainWindowRect = GUILayout.Window(mainGuid, MainWindowRect, MainGUI, "Persistent Rotation");
                    }
                    if (showBodyWindow)
                    {
                        BodyWindowRect = GUILayout.Window(bodyGuid, BodyWindowRect, BodyGUI, "Select Body");
                    }
                    if (showConfigWindow)
                    {
                        ConfigWindowRect = GUILayout.Window(configGuid, ConfigWindowRect, ConfigGUI, "Configure");
                    }
                    if (showGravityWindow)
                    {
                        GravityWindowRect = GUILayout.Window(gravityGuid, GravityWindowRect, GravityGUI, "Artifical Gravity");
                    }
                }
            }
        }
        void OnShowUI()
        {
            hidden = false;
        }
        void OnHideUI()
        {
            hidden = true;
        }
        private void MainGUI(int windowID)
        {
            Texture toggle;

            if (MainWindowActive)
            {
                toggle = GameDatabase.Instance.GetTexture("PersistentRotation/Textures/minimize_w", false);
            }
            else
            {
                toggle = GameDatabase.Instance.GetTexture("PersistentRotation/Textures/maximize_w", false);
            }

            GUILayout.BeginVertical();
            GUILayout.BeginHorizontal();
            GUILayout.FlexibleSpace();
            if (GUILayout.Button(options))
            {
                showConfigWindow = !showConfigWindow;
            }
            if (GUILayout.Button(toggle))
            {
                MainWindowActive = !MainWindowActive;
                MainWindowRect = new Rect(MainWindowRect.x, MainWindowRect.y, 200, 10);
            }
            GUILayout.EndHorizontal();
            if (MainWindowActive)
            {
                if (GUILayout.Button("Body Relative Rotation", GUILayout.ExpandWidth(true)))
                {
                    showBodyWindow = !showBodyWindow;
                }
                if (vessel_reference.ContainsKey(FlightGlobals.ActiveVessel.id.ToString()))
                {
                    if (vessel_reference[FlightGlobals.ActiveVessel.id.ToString()] != null)
                    {
                        GUILayout.Label("Current Reference: " + vessel_reference[FlightGlobals.ActiveVessel.id.ToString()].GetName());
                    }
                    else
                    {
                        GUILayout.Label("Current Body: none");
                    }
                }
                else
                {
                    GUILayout.Label("Current Body: none");
                }
                GUILayout.Space(20);
                if (GUILayout.Button("Artificial Gravity", GUILayout.ExpandWidth(true)))
                {
                    showGravityWindow = !showGravityWindow;
                }
            }
            GUILayout.EndVertical();
            GUI.DragWindow();

        }
        private void BodyGUI(int windowID)
        {
            GUILayout.BeginVertical();
            GUILayout.BeginHorizontal();
            GUILayout.FlexibleSpace();
            if (GUILayout.Button(close))
            {
                showBodyWindow = false;
            }
            GUILayout.EndHorizontal();
            GUILayout.Label("Select a target in map view\nand press <Set> to use it as reference.");
            if (GUILayout.Button("Set", GUILayout.ExpandWidth(true)))
            {
                if(FlightGlobals.ActiveVessel.targetObject.GetType() == typeof(CelestialBody) || FlightGlobals.ActiveVessel.targetObject.GetType() == typeof(Vessel))
                    vessel_reference[FlightGlobals.ActiveVessel.id.ToString()] = FlightGlobals.ActiveVessel.targetObject;
            }
            if (GUILayout.Button("Unset", GUILayout.ExpandWidth(true)))
            {
                vessel_reference[FlightGlobals.ActiveVessel.id.ToString()] = null;

            }
            GUILayout.Space(10);
            if (GUILayout.Button("Sun", GUILayout.ExpandWidth(true)))
            {
                vessel_reference[FlightGlobals.ActiveVessel.id.ToString()] = Sun.Instance.sun;
            }
            if (GUILayout.Button(FlightGlobals.ActiveVessel.mainBody.name, GUILayout.ExpandWidth(true)))
            {
                vessel_reference[FlightGlobals.ActiveVessel.id.ToString()] = FlightGlobals.ActiveVessel.mainBody;
            }

            GUILayout.EndVertical();
            GUI.DragWindow();
        }
        private void ConfigGUI(int windowID)
        {
            GUILayout.BeginVertical();
            GUILayout.BeginHorizontal();
            GUILayout.FlexibleSpace();
            if(GUILayout.Button(close))
            {
                showConfigWindow = false;
            }
            GUILayout.EndHorizontal();
            GUILayout.Label("Configure Visibility");
            if (GUILayout.Button("Always Visible", GUILayout.ExpandWidth(true)))
            {
                visibility_mode = 1;
                DeleteBlizzyToolbar();
                DeleteStockToolbar();
            }
            if (GUILayout.Button("Stock Toolbar", GUILayout.ExpandWidth(true)))
            {
                visibility_mode = 2;
                CreateStockToolbar();
                DeleteBlizzyToolbar();
            }
            if (ToolbarManager.ToolbarAvailable)
            {
                if (GUILayout.Button("Blizzy's Toolbar", GUILayout.ExpandWidth(true)))
                {
                    visibility_mode = 3;
                    CreateBlizzyToolbar();
                    DeleteStockToolbar();
                }
            }
            GUILayout.EndVertical();
            GUI.DragWindow();
        }
        private void GravityGUI(int windowID)
        {
            GUILayout.BeginVertical();
            GUILayout.BeginHorizontal();
            GUILayout.FlexibleSpace();
            if (GUILayout.Button(close))
            {
                showGravityWindow = false;
            }
            GUILayout.EndHorizontal();
            if (lastVessel.GetCrewCount() > 0 && lastVessel.vesselType != VesselType.EVA)
            {
                foreach (KeyValuePair<ProtoCrewMember, float> entry in ArtificialGravity)
                {
                    GUILayout.BeginHorizontal();
                    GUILayout.Label(entry.Key.name);
                    GUILayout.FlexibleSpace();
                    GUILayout.Label(((double)((int)((entry.Value / 9.81) * 10.0)) / 10.0).ToString() + " G");
                    GUILayout.EndHorizontal();
                }
            }
            else
            {
                GUILayout.Label("No crew onboard.");
            }
            GUILayout.EndVertical();
            GUI.DragWindow();
        }
        void CreateBlizzyToolbar()
        {
            if (ToolbarManager.ToolbarAvailable)
            {
                // regular button
                button = ToolbarManager.Instance.add("PersistentRotation", "button");
                button.TexturePath = "PersistentRotation/Textures/texture";
                button.ToolTip = "Toggle PersistentRotation";
                button.OnClick += (e) => 
                {
                    disabled = !disabled;
                };
            }
        }
        void DeleteBlizzyToolbar()
        {
            if (ToolbarManager.ToolbarAvailable)
            {
                if (button != null)
                {
                    button.Destroy();
                }
            }
        }
        void CreateStockToolbar()
        {
            if (!stockButton)
            {
                    stockButton = ApplicationLauncher.Instance.AddModApplication(
                    () =>
                    {
                        disabled = !disabled;
                    },
                    () =>
                    {
                        disabled = !disabled;
                    },
                    null,
                    null,
                    null,
                    null,
                    ApplicationLauncher.AppScenes.FLIGHT,
                    GameDatabase.Instance.GetTexture("PersistentRotation/Textures/texture", false)
                    );
            }
        }
        void DeleteStockToolbar()
        {
            if (stockButton)
            {
                ApplicationLauncher.Instance.RemoveModApplication(stockButton);
                stockButton = null;
            }
        }
        void SaveGUI()
        {
            try
            {
                ConfigNode save = new ConfigNode();
                ConfigNode.CreateConfigFromObject(this, 0, save);

                save.AddValue("disabled", disabled.ToString());
                save.AddValue("mode", visibility_mode.ToString());
                save.AddValue("active", MainWindowActive.ToString());

                save.AddValue("showMain", showMainWindow.ToString());
                save.AddValue("xMain", MainWindowRect.x.ToString());
                save.AddValue("yMain", MainWindowRect.y.ToString());

                save.AddValue("showBody", showBodyWindow.ToString());
                save.AddValue("xBody", BodyWindowRect.x.ToString());
                save.AddValue("yBody", BodyWindowRect.y.ToString());

                save.AddValue("showConfig", showConfigWindow.ToString());
                save.AddValue("xConfig", ConfigWindowRect.x.ToString());
                save.AddValue("yConfig", ConfigWindowRect.y.ToString());

                save.AddValue("showGravity", showGravityWindow.ToString());
                save.AddValue("xGravity", GravityWindowRect.x.ToString());
                save.AddValue("yGravity", GravityWindowRect.y.ToString());

                save.Save(gui_cfg);
            }
            catch (Exception e) { Debug.Log("[PR] Saving not successful: " + e.Message); }
        }
        void LoadGUI()
        {
            ConfigNode load = ConfigNode.Load(gui_cfg);
            if (load == null)
            {
                Debug.Log("[PR] Cfg file is empty or not existent!");
                return;
            }

            foreach(ConfigNode.Value s in load.values)
            {
                if (s.name == "disabled")
                    disabled = Convert.ToBoolean(s.value);
                if (s.name == "mode")
                    visibility_mode = Convert.ToInt32(s.value);
                if (s.name == "active")
                    MainWindowActive = Convert.ToBoolean(s.value);

                if (s.name == "showMain")
                    showMainWindow = Convert.ToBoolean(s.value);
                if (s.name == "xMain")
                    MainWindowRect.x = Convert.ToSingle(s.value);
                if (s.name == "yMain")
                    MainWindowRect.y = Convert.ToSingle(s.value);

                if (s.name == "showBody")
                    showBodyWindow = Convert.ToBoolean(s.value);
                if (s.name == "xBody")
                    BodyWindowRect.x = Convert.ToSingle(s.value);
                if (s.name == "yBody")
                    BodyWindowRect.y = Convert.ToSingle(s.value);

                if (s.name == "showConfig")
                    showConfigWindow = Convert.ToBoolean(s.value);
                if (s.name == "xConfig")
                    ConfigWindowRect.x = Convert.ToSingle(s.value);
                if (s.name == "yConfig")
                    ConfigWindowRect.y = Convert.ToSingle(s.value);

                if (s.name == "showGravity")
                    showGravityWindow = Convert.ToBoolean(s.value);
                if (s.name == "xGravity")
                    GravityWindowRect.x = Convert.ToSingle(s.value);
                if (s.name == "yGravity")
                    GravityWindowRect.y = Convert.ToSingle(s.value);
            }
        }
    }
}